//
//  新規追加したコマンド
//   +0      +1      +2        +3         +35
//  [ID] [PAGE_TX_*] [len=32] [data(32)] [ isp(4) ]

//
//  変更されたコマンド番号
//      HIDASP_PAGE_TX 20 ===> 0x20
//      HIDASP_PAGE_RD 21 ===> 0x21
//

//
//  仕様変更されたコマンド
//  HIDASP_TEST 0x01
//    旧仕様:    [ID] [0x01] [arg1]  ====> [ID] [arg1] をエコーバック.
//    新仕様:    [ID] [0x01] [arg1]  ====> [ID] [DEV_ID] [arg1] をエコーバック.
//
//  DEV_ID
//      0x55 : フュージョンありファームウェア
//      0x5A : フュージョンなしファームウェア
//

#define HIDASP_TEST           0x01
#define HIDASP_SET_STATUS     0x02
#define HIDASP_GET_CAPS       0x03
#define HIDASP_CMD_TX         0x10
#define HIDASP_CMD_TX_1       0x11
#define HIDASP_SET_PAGE       0x14
#define HIDASP_PAGE_TX        0x20  //MODE=00
#define HIDASP_PAGE_RD        0x21  //MODE=01
#define HIDASP_PAGE_TX_START  0x22  //MODE=02  ADDR 初期化.
#define HIDASP_PAGE_TX_FLUSH  0x24  //MODE=04  転送後 ISP
//                            0x27 まではPAGE_TXで使用する.

#define CMD_MASK              0xf8
#define MODE_MASK             0x07

#define HIDASP_BOOT_EXIT	  0x36  //bootloadを終了し、アプリケーションを起動する.
#define HIDASP_BOOT_RWW	      0x38  //アプリケーション領域のコード読み出しを許可する.
#define HIDASP_JMP	          0x3a  //指定番地のプログラム実行.
#define HIDASP_SET_DELAY      0x3c  //SPI書き込み遅延時間の設定.
#define HIDASP_POKE           0x3e  //メモリー書き込み.
#define HIDASP_PEEK           0x3f  //メモリー読み出し.

#define	DEV_ID_FUSION			0x55
#define	DEV_ID_STD				0x5a
#define	DEV_ID_MEGA88			0x88
#define	DEV_ID_MEGA88_USERMODE	0x89

/*
 * ドキュメント

  +--------------+-------------------------------------------------------------+
  |   コマンド   |  パラメータ              | 動作                             |
  +--------------+-----------+-------------------------------------------------+
  |TEST          | ECHO      |          ＝接続チェックを行う.
  +--------------+-----------+--------+----------------------------------------+
  |SET_STATUS    | PORTD     | PORTB  | ＝LED制御を行う.
  +--------------+-----------+--------+----------------------------------------+
  |CMD_TX        | ISP_COMMAND[4]           | ＝ISPコマンドを４バイト送信
  +--------------+--------------------------+----------------------------------+
  |CMD_TX_1      | ISP_COMMAND[4]           | ＝ISPコマンドを４バイト送信して受信
  +--------------+-----------+--------------+----------------------------------+
  |SET_PAGE      | PAGE_MODE |   ＝ページモードの初期化とアドレスリセット
  +--------------+-----------+-------------------------------------------------+
  |PAGE_TX       | LENGTH    |   DATA[ max 32byte ]       | ＝ページモード書き込み
  +--------------+-----------+-------------------------------------------------+
  |PAGE_RD       | LENGTH    |                              ＝ページモード読み出し
  +--------------+-----------+-------------------------------------------------+
  |PAGE_TX_START | LENGTH    |   DATA[ max 32byte ]       | ＝ページモードの初期化とページモード書き込み
  +--------------+-----------+-------------------------------------------------+
  |PAGE_TX_FLUSH | LENGTH    |   DATA[ 32byte 固定]       | ISP_COMMAND[4]    | ＝ページモード書き込み後Ｆｌｕｓｈ書き込み実行.
  +--------------+-----------+-------------------------------------------------+
  |SET_DELAY     | delay     |   ＝ＳＰＩクロック速度の設定(delay=0..255).
  +--------------+-----------+------------+-------+-------+--------------------+
  |POKE          | count     | address[2] | data0 | data1 | ＝内蔵ＲＡＭ書き込み
  +--------------+-----------+------------+-------+-------+--------------------+
  |PEEK          | count     | address[2] |                 ＝内蔵ＲＡＭ読み出し
  +--------------+-----------+-------------------------------------------------+

 *注1: PAGE_TX_STARTとPAGE_TX_FLUSHはどちらもPAGE_TXに対する修飾ビットなので、
       両方を併せ持つパケットも作成可能です。

 *注2: コマンドのLSBが１になっている場合はコマンド実行後、読み取りフェーズが実行されます。
       読み取りサイズは、ホスト側から指定されます。

  応答フェーズで返却されるパケットの構造
  +--------------+-------------------------------------------------------------+
  |   コマンド   |  返却値                                                     |
  +--------------+-----------+----------+--------------------------------------+
  |TEST          | DEV_ID    | ECHO     |           ＝接続チェック.
  +--------------+-----------+----------+---+----------------------------------+
  |CMD_TX_1      | ISP_COMMAND[4]           | ＝ISPコマンド実行結果.
  +--------------+--------------------------+----------------------------------+
  |PAGE_RD       | DATA[ max 38byte ]       | ＝ページモード読み出しデータ.
  +--------------+--------------------------+----------------------------------+
  |PEEK          | DATA[ max 8 byte ]       | ＝内蔵ＲＡＭ読み出しデータ.
  +--------------+--------------------------+----------------------------------+

 *注3: HID Reportは先頭に必ずReportIDが入るため、
       上記返却値は HID Reportの２バイト目以降の意味となります。

 *注4: DEV_ID は現在の版では、以下の値のどちらかです。 
		0x55 : フュージョンありファームウェア
		0x5A : フュージョンなしファームウェア
       ECHOは渡された値をそのまま返します.


*/
